;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")


(defun pick-parsed-atm (number parsed-atm-vector)
  "number             ;indicates the atm to be returned
   parsed-atm-vector
Scans the parsed-atm-vector and returns the atm-struct which corresponds
to the parsed-atm-number"
  (let ((result-atm))
    (dosequence (parsed-atm parsed-atm-vector result-atm)
                (unless (eql parsed-atm '())
                  (if (= (get-parsed-atm-number parsed-atm) number)
                    (progn (setf result-atm (get-parsed-atm parsed-atm))
                           (return))
                    )))
    result-atm))

(defun create-link-type (&key parent-name
                                 parent-site
                                 parent-bonds
                                 child-name
                                 child-site
                                 raw-format
                                 parent-coord-atms
                                 child-coord-atms
                                 (clash-atms '())
                                 (lk-name (concatenate
                                           'string
                                           parent-name "->" child-name)))
  (let ((parent-exists? (retrieve-from-mlib parent-name))
        (child-exists? (retrieve-from-mlib child-name)))
    (if (and parent-exists? child-exists?)
      (let* ((parsed-atm-vector (first raw-format))
             (parent-points (list (atm-xyz (pick-parsed-atm
                                            (first parent-coord-atms)
                                            parsed-atm-vector))
                                  (atm-xyz (pick-parsed-atm
                                            (second parent-coord-atms)
                                            parsed-atm-vector))
                                  (atm-xyz (pick-parsed-atm
                                            (third parent-coord-atms)
                                            parsed-atm-vector))))
             (child-points (list (atm-xyz (pick-parsed-atm
                                           (first child-coord-atms)
                                           parsed-atm-vector))
                                 (atm-xyz (pick-parsed-atm
                                           (second child-coord-atms)
                                           parsed-atm-vector))
                                 (atm-xyz (pick-parsed-atm
                                           (third child-coord-atms)
                                           parsed-atm-vector))))
             (child-bonds (invert-bond-relation parent-bonds))
             (the-link-type
              (make-link-type
               ;; :number   **gets assigned at the time of entry into the mlib** 
               :parent       parent-name
               :parent-site  parent-site
               :parent-bonds parent-bonds
               :child        child-name
               :child-site   child-site
               :child-bonds  child-bonds
               :name         lk-name
               :matrix       (mapping-matrix parent-points child-points)
               :volume         10
               :clash-atms   clash-atms
               )))
        the-link-type)
      ;;if any of the moieties do not exist in the library:
      (progn (if (not parent-exists?)
               (pprint (concatenate 'string
                                    "the parent moiety-type : "
                                    parent-name
                                    " was not found in the mlib")))
             (if (not child-exists?)
               (pprint (concatenate 'string
                                    "the child moiety-type : "
                                    child-name
                                    " was not found in the mlib"))))
      )))

(defun invert-bond-relation (bond-relation)
  "bond-relation    ;list of bond-starlets
Takes a list of bond-starlets which describes bonds from, say, a parent
to a child moiety, and returns a list of bond-starlets that describes
the same bonds, but with inverted perspective, i.e. looking now from
child to the parent. This inversion involves making the former bond-ends
be the new center-atms."
  (let ((result-bond-relation '()))
    (dosequence (bond-starlet bond-relation result-bond-relation)
                (let ((old-center-atm (get-center-atm bond-starlet)))
                  (dosequence (old-end-atm (get-bond-ends bond-starlet))
                              (let ((already-exists-in-result
                                     (position-if
                                      #'(lambda (starlet)
                                          (eql old-end-atm
                                               (get-center-atm starlet)))
                                      result-bond-relation)))
                                (if (eql already-exists-in-result nil)
                                  (push (list old-end-atm
                                              (list old-center-atm))
                                        result-bond-relation)
                                  (let* ((temp
                                          (elt result-bond-relation
                                               already-exists-in-result))
                                         (bond-end-list
                                          (get-bond-ends temp)))
                                    (pushnew old-center-atm bond-end-list)
                                    (setf 
                                     (elt result-bond-relation
                                          already-exists-in-result)
                                     (list (get-center-atm temp)
                                           bond-end-list))))))))))
#|
(invert-bond-relation '((32 (47))))
(invert-bond-relation '((1 (1 2))))
(invert-bond-relation '((2 (1)) (1 (1))))
(invert-bond-relation '((32 (1 2 3 4)) (17 (1 5 6 7)) (16 (89 1))))
(invert-bond-relation
 '((89 (16)) (7 (17)) (6 (17)) (5 (17)) (4 (32)) (3 (32)) (2 (32))
   (1 (16 17 32)))
)
|#

;;; This function is called in add-link-type-to-mlib
(defun enter-link-type-into-parent-site (link-type)
  "link-type
Enters this link-type into the parent-moiety-type's site structure, into
the :link-types list which is ordered according to decreasing volume."
  (let* ((parent-moiety (retrieve-from-mlib (link-type-parent link-type)))
         (parent-site (elt (moiety-type-sites parent-moiety)
                           (link-type-parent-site link-type)))
         (link-list (site-link-types parent-site)))
    (push (link-type-name link-type) link-list)
    (setf (site-link-types parent-site)
          (stable-sort link-list
                       #'(lambda (a b)
                           (> (link-type-volume (retrieve-from-mlib a))
                              (link-type-volume (retrieve-from-mlib b))))))
    ))

;;; This function is called in delete-link-type-from-mlib
(defun delete-link-type-in-parent-site (link-type)
  "link-type
Deletes this link-type from the parent-moiety-type's site structure, from
the :link-types list."
  (let* ((parent-moiety (retrieve-from-mlib (link-type-parent link-type)))
         (parent-site (elt (moiety-type-sites parent-moiety)
                           (link-type-parent-site link-type)))
         (link-list (site-link-types parent-site)))
    (setf (site-link-types parent-site)
          (delete (link-type-name link-type) link-list))
    ))

;;; This function is called in delete-from-mlib
(defun alarm-link-types-if-parent-deleted (moiety-type)
  "moiety-type
Whenever a moiety-type is supposed to be deleted from the mlib, calling
this routine will produce an alarm message showing all the link-types that
will become dangling."
  (pprint (concatenate 'string
                       "ALARM: by deleting moiety-type : "
                       (moiety-type-name moiety-type)
                       " the following link-types will dangle :"))
  (let ((site-vector (moiety-type-sites moiety-type)))
    (dotimes (site-number (length site-vector))
      (pprint (concatenate 'string
                           "SITE "
                           (write-to-string site-number)))
      (pprint (concatenate 'string
                           (write-to-string
                            (site-link-types
                             (elt site-vector site-number)))))
      )))

