;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; This is a first, primitive tree grower which drives the low-level routines.
;;; The "heuristic" used for selecting the next attachment site and the next
;;; moiety is simply provided by a random generator.
;;; A new data structure called "open-sites", which is just a list of so-called
;;; "tree-sites", is introduced to keep track of all the 'open sites on the growing
;;; tree structure. It is easy to randomly pick a new attachment site from this
;;; linear list. This data structure gets initialized by being called up from
;;; (init-tree frame-moiety).

(defstruct tree-site
  parent        ; pointer to parent moiety
  site          ; site-index
  link-types    ; list of possible link types for site,
                ;   ordered in the same way as in the moiety-type
  )

(defun init-open-sites ()
  "no parameters
Initializes the list which is an unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (setf (get (get-current-tree-symbol) 'open-sites) '()))

(defun get-open-sites ()
  "no parameters
Obtains the list which is an unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (get (get-current-tree-symbol) 'open-sites))

(defun number-of-open-sites ()
  "no parameters
Obtains the number of the currently 'open sites
of the stuffing-tree which is under construction."
  (length (get-open-sites)))

(defun set-open-sites (list)
  "list
Puts in place a new list which is an unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (setf (get (get-current-tree-symbol) 'open-sites) list))

(defun get-from-open-sites (elt-index)
  "elt-index     ; element-index of the list (usually quite random)
Obtains the indicated element from the unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (elt (get-open-sites) elt-index))

(defun remove-from-open-sites (tree-site)
  "tree-site
Removes the indicated element from the unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (set-open-sites
   (delete tree-site (get-open-sites))))

(defun add-to-open-sites (tree-site)
  "tree-site
Adds the indicated site in the unordered set of the currently 'open sites
of the stuffing-tree which is under construction."
  (push tree-site (get (get-current-tree-symbol) 'open-sites)))

(defun find-open-tree-site (parent site-index)
  "parent
   site-index
"
  (find-if #'(lambda (tree-site)
               (and (eql (tree-site-parent tree-site) parent)
                    (eql (tree-site-site tree-site) site-index)))
           (get-open-sites)))


(defun enlist-open-sites (moiety)
  "moiety
Enlists all the 'open sites of a (newly added) moiety in the unordered set of the
currently 'open sites of the stuffing-tree which is under construction."
  (let ((sites (moiety-site-status moiety)))
    (dotimes (site-index (length sites) (values))
      (when (eql (elt sites site-index) 'open)
        (let ((tree-site (make-tree-site
                          :parent      moiety
                          :site        site-index
                          :link-types  (site-link-types
                                        (elt (moiety-type-sites
                                              (moiety-moiety-type moiety))
                                             site-index))
                          )))
          (add-to-open-sites tree-site))))))

(defun manage-site-status (moiety site-index new-status)
  "moiety
   site-index
   new-status
Change the site-status to new-status and accordingly enter or delete this site
into the open-sites data structure."
  (let ((previous-status (elt (moiety-site-status moiety) site-index)))
    (setf (elt (moiety-site-status moiety) site-index) new-status)
    ;; now manage the open-sites data structure:
    (ecase previous-status
           (open (when (eql new-status 'used)     ;; site has been consumed during
                    (remove-from-open-sites       ;;   tree construction
                     (find-open-tree-site moiety site-index)))
                  (when (eql new-status 'blocked) ;; hope for attaching something
                    (remove-from-open-sites       ;;  here has been given up
                     (find-open-tree-site moiety site-index))))
           ((used blocked) (when (eql new-status 'open)
                               (add-to-open-sites
                                (make-tree-site
                                 :parent      moiety
                                 :site        site-index
                                 :link-types  (site-link-types
                                               (elt (moiety-type-sites
                                                     (moiety-moiety-type moiety))
                                                    site-index))
                                 )))))
    ))

(defun remove-link-from-tree-site (tree-site link-type-indicator)
  "tree-site
   link-type-indicator
Removes a link-type from the list of possible link-types of a site in the tree.
This can be done e.g. after a clash has been detected."
  (setf (tree-site-link-types tree-site)
        (delete link-type-indicator (tree-site-link-types tree-site)
                :test #'equal)))

(defun grow-one-moiety (next-tree-site)
  "next-tree-site
At a given ('open) tree-site, this routine randomly picks a link-type and tests
for a clash. If it clashes, the link-type gets removed from the link-type-entries
of the tree-site.
If there is no clash, the moiety is added to the tree (implicitly
consuming this tree-site and turning it into 'used)."
  (let* ((parent (tree-site-parent next-tree-site))
         (parent-site-index (tree-site-site next-tree-site))
         (link-type-indicator (random-pick (tree-site-link-types next-tree-site)))
         (link-type (retrieve-from-mlib link-type-indicator)))
    (if (attach-if-no-clash-p parent parent-site-index link-type)
      ;; has clashed
      (progn
        (remove-link-from-tree-site next-tree-site link-type-indicator)
        (if (eql '() (tree-site-link-types next-tree-site))
          ;; no link-types remain for this site, which must mean that
          ;;   all of them are sterically blocked
          (manage-site-status parent parent-site-index 'blocked)  ;; removes site
          ;; there still are link-types to choose from
          nil)
        )
      ;; no clash (moiety has been attached)
      nil)))

(defun grow-a-tree ()
  "no parameters
"
  (do ()
      ((= 0 (number-of-open-sites)) 'tree-finished)   ;; termination condition
    (let ((next-tree-site (random-pick (get-open-sites))))
      (grow-one-moiety next-tree-site))
    ))


(defvar Current-tree-rack nil
  "Denotes the currently active rack of stuffing-trees in their shorthand form.")

(defun get-tree-rack-symbol ()
  "no parameters
Obtains the symbol which harbors the currently active rack of stuffing-trees."
  (values Current-tree-rack))

(defun get-tree-rack ()
  "no parameters
Obtains the currently active rack of stuffing-trees."
  (get (get-tree-rack-symbol) 'tree-rack))

(defun set-tree-rack (rack)
  "rack     ;essentially a list
Sets the currently active rack of stuffing-trees to the supplied one."
  (setf (get (get-tree-rack-symbol) 'tree-rack) rack))

(defun init-tree-rack ()
  "no parameters
"
  (set-tree-rack '())
  (set-tree-rack-max-size 15))

(defun get-tree-rack-max-size ()
  "no parameters
Gets the maximum number of stuffing-trees to be retained."
  (get (get-tree-rack-symbol) 'max-size))

(defun set-tree-rack-max-size (size)
  "size
Sets the maximum number of stuffing-trees to be retained."
  (setf (get (get-tree-rack-symbol) 'max-size) size))

(defun volume-sort-add-to-tree-rack (shorthand-tree volume)
  "shorthand-tree
   volume           ;of the tree to be added
Inserts a stuffing tree in its space-saving shorthand representation into the 
volume-ordered tree-rack, while not allowing the rack to grow larger than a 
predetermined size. This means that the worst case tree at the very end of the 
rack will just fall off into oblivion."
  (let* ((tree-with-vol (push (list 'volume volume) shorthand-tree)) ;attach volume
         (temp-rack (stable-sort (append (get-tree-rack) (list tree-with-vol))
                                 #'(lambda (larger smaller)
                                     (< (second (first smaller)) ;: volume
                                        (second (first larger))) ;: volume
                                     )
                                 ))
         (max-size (get-tree-rack-max-size)))
    (when (> (length temp-rack)       ;; truncate if necessary
             max-size)
      (setf temp-rack (subseq temp-rack 0 max-size)))
    (set-tree-rack temp-rack)))

(defun tree-rack-volumes ()
  "no parameters
Returns a list of the volumes that those stored stuffing-trees are claiming to 
be able to fill."
  (map 'list #'(lambda (shorthand-tree)
                 (second (first shorthand-tree))) ;: volume
       (get-tree-rack)))

(defun grow-a-bunch-of-trees (number-of-iterations)
  (init-tree-rack)
  (dotimes (count number-of-iterations)
    (init-tree "Cubane-frame")
    (grow-a-tree)
    (volume-sort-add-to-tree-rack (dendro-stenogen (get-current-tree))
                                  (get-current-tree-volume))
    (print "Just finished tree number:")  ; with preceding newline
    (prin1 count)
    (print "has volume of:")              ; with preceding newline
    (prin1 (get-current-tree-volume))
    ))

(defun save-top-ten ()
  (let ((INCLUDE-FRAME t)
        (NO-INCLUDE-FRAME nil)
        (top-ten (subseq (get-tree-rack) 0 10)))
    (declare (ignore NO-INCLUDE-FRAME))
    (doseq-with-index (shorthand-tree index top-ten)
      (dendro-plerogen (cdr shorthand-tree)) ;; leave away volume information
      (write-pdb-file (dump-tree (get-current-tree) INCLUDE-FRAME) 
                      (concatenate 'string
                                   "tree"
                                   (write-to-string index)
                                   ".pdb"))
      )))
