;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Historical Note: the first version of this file was made by K. E. Drexler.
;;;                  subsequent modifications by KR.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;;
;;; Code for matrix and vector manipulations:
;;;
;;; Where dx = offset in x, dy = offset in y, etc.
;;; xx = projection of x unit vector on new x axis
;;; yx = projection of y unit vector on new x axis, etc.
;;; A transformation matrix is then:
;;; #2A((xx   yx   zx  dx)
;;;     (xy   yy   zy  dy)
;;;     (xz   yz   zz  dz)
;;;     ( 0    0    0   1))
;;; And the corresponding vector format is:
;;;    #(x y z 1)
;;; The effect of multiplication is rotation + offset
;;;
;;; Specialized rotation-displacement matrices:
;;; Drop the information-free row, yielding
;;; #2A((xx   yx   zx  dx)
;;;     (xy   yy   zy  dy)
;;;     (xz   yz   zz  dz))
;;; And:
;;;    #(x y z)

;;; Make an empty matrix of the above kind.
(defun new-matrix ()
  "no parameters
Make an empty special-matrix of elements of type 'real"
  (make-array '(3 4) :element-type 'real))
#|
(new-matrix)
|#

;;; Make an empty vector of the above kind.
(defun new-vector ()
  "no parameters
Make an empty vector of elements of type 'real"
  (make-array 3 :element-type 'real))
#|
(new-vector)
|#

;;; Make a unit matrix of the above kind.
(defun new-unit-matrix ()
  "no parameters
Make a pre-initialized special-matrix of elements of type 'real,
containing the unit transform."
  (make-array '(3 4) :element-type 'real
                     :initial-contents
                     '((1.0  0.0  0.0  0.0)
                       (0.0  1.0  0.0  0.0)
                       (0.0  0.0  1.0  0.0))))
#|
(new-unit-matrix)
|#

;;; For these special-matrices, multiply
;;; a vector by a matrix.
;;; (Mnemonic: mv* = matrix vector multiply)
(defun mv* (matrix vector)
  "matrix vector
Multiply a vector by a special-matrix"
  (let ((result-vector (new-vector)))
    (dotimes (j 3)
      (let ((rowsum (aref matrix j 3)))
        (dotimes (i 3)
          (incf rowsum (* (aref vector i)
                          (aref matrix j i))))
        (setf (aref result-vector j) rowsum)))
    result-vector))
#|
(time
 (dotimes (i 1000)
   (mv* #2A((1.0  0.0  0.0  1.0)
            (0.0  1.0  0.0  2.0)
            (0.0  0.0  1.0  3.0))
        #(1.0  1.0  1.0))
   ))
|#

;;; #2A((xx   yx   zx  dx)
;;;     (xy   yy   zy  dy)
;;;     (xz   yz   zz  dz)
;;;     ( 0    0    0   1))

;;; Multiply MM, where Ms are as above.
;;; (mm* matrix matrix multiply)
(defun mm* (left-matrix right-matrix)
  "left-matrix right-matrix
Multiply a special-matrix with a special-matrix"
  (let ((result-matrix (new-matrix)))
    (dotimes (k 3)
      (dotimes (j 3)
        (let ((rowsum 0.0))
          (dotimes (i 3)
            (incf rowsum (* (aref right-matrix i k)
                            (aref left-matrix j i))))
          (setf (aref result-matrix j k) rowsum))))
    (dotimes (j 3)
      (let ((rowsum (aref left-matrix j 3)))
        (dotimes (i 3)
          (incf rowsum (* (aref right-matrix i 3)
                          (aref left-matrix j i))))
        (setf (aref result-matrix j 3) rowsum)))
    result-matrix))
#|
;; not rotation matrices, but same form:
(time
 (dotimes (i 1000)
   (mm* #2A((1.0  0.0  10.0  1.0)
            (0.0  1.0  0.0  1.0)
            (0.0  0.0  1.0  1.0))
        #2A((1.0  0.0  3.0  0.0)
            (0.0  2.0  0.0  0.0)
            (1.0  0.0  3.0  0.0))
        )
   ))
;; test of associativity:
(let ((a #2A((8.0  4.0  3.0  4.0)
             (6.0  2.0  3.0  2.0)
             (2.0  7.0  7.0  7.0)))
      (b #2A((6.0  3.0  2.0  4.0)
             (3.0  1.0  9.0  2.0)
             (7.0  6.0  3.0  5.0)))
      (c #(1.0 2.0 3.0)))
  (print (mv* (mm* a b) c))
         (mv* a (mv* b c)))
|#

;;; Form the dot product of two 3-vectors.
(defun dot (vector1 vector2)
  "vector1 vector2
Form the dot product of two vectors"
  (let ((result 0.0))
    (dotimes (i 3)
      (incf result
            (* (aref vector1 i)
               (aref vector2 i))))
  result))
#|
(dot #(1 2 3) #(1 1 1))
|#

;;; Multiply a vector by a scalar.
(defun sv* (scalar vector)
  "scalar vector
Multiply a vector by a scalar"
  (let ((result-vector (new-vector)))
    (dotimes (i 3)
      (setf (aref result-vector i)
            (* scalar (aref vector i))))
    result-vector))

#|
(sv* 5 #(1 1 1))
|#

;;; Multiply a matrix by a scalar.
(defun sm* (scalar matrix)
  "scalar matrix
Multiply a matrix by a scalar"
  (let ((result-matrix (new-matrix)))
    (dotimes (i 4)
      (dotimes (j 3)
        (setf (aref result-matrix j i)
              (* scalar (aref matrix j i)))))
  result-matrix))
#|
(sm* 5 #2A((1.0 0.0 1.0 0.0)
           (0.0 1.0 0.0 0.0)
           (1.0 0.0 1.0 0.0)))
|#

;;; Subtract a vector from a vector.
(defun vv- (vector1 vector2)
  "vector1 vector2
Subtract vector2 from vector1"
  (let ((result-vector (new-vector)))
    (dotimes (i 3)
      (setf (aref result-vector i)
            (- (aref vector1 i) (aref vector2 i))))
    result-vector))
#|
(vv- #(1 1 1) #(1 1 0))
|#

;;; Add a vector to a vector.
(defun vv+ (vector1 vector2)
  "vector1 vector2
Add a vector to a vector"
  (let ((result-vector (new-vector)))
    (dotimes (i 3)
      (setf (aref result-vector i)
            (+ (aref vector1 i) (aref vector2 i))))
    result-vector))
#|
(vv+ #(1 1 1) #(1 1 0))
|#


;;; Normalize a 3-vector, returning a unit vector.
(defun unit (vector)
  "vector
Normalize a vector, returning a unit vector"
  (let ((result-vector (new-vector))
        (divisor (sqrt (dot vector vector))))
    (if (= divisor 0.0) (error "cannot make unit vector from 0-vector"))
    (dotimes (i 3)
      (setf (aref result-vector i)
            (/ (aref vector i) divisor)))
  result-vector))
#|
(unit #(5 5 0))
|#

;;; Form the cross product of two 3-vectors.
(defun cross (vector1 vector2)
  "vector1 vector2
Form the cross product of two vectors"
  (let ((result-vector (new-vector)))
    (dotimes (i 3)
      (setf (aref result-vector i)
            (- (* (aref vector1 (mod (+ i 1) 3))
                  (aref vector2 (mod (+ i 2) 3)))
               (* (aref vector1 (mod (- i 1) 3))
                  (aref vector2 (mod (- i 2) 3))))))
    result-vector))
#|
(cross #(0 0 1) #(1 0 0))
|#

;;; Find the length of a vector.
(defun v-length (vector)
  "vector
Find the length of a vector"
  (let ((sum-squares 0.0))
    (dotimes (i 3)
      (incf sum-squares
            (* (aref vector i)
               (aref vector i))))
    (sqrt sum-squares)))
#|
(v-length #(0 1 1))
|#

;;; Find the distance between two points.
(defun v-distance (vector1 vector2)
  "vector1 vector2
Find the distance between two points"
  (v-length (vv- vector2 vector1)))
#|
(v-distance #(0 1 0) #(1 0 0))
|#

;;; Make a unit vector orthogonal to a reference
;;; vector, in the plane of a second vector and
;;; having a positive dot product with it.
(defun orthogonal-unit (reference-vector second-vector)
  "reference-vector second-vector
Make a unit vector orthogonal to a reference
vector, in the plane of a second vector and
having a positive dot product with it"
  (let* ((unit-ref (unit reference-vector))
         (projection (dot unit-ref second-vector))
         (orthogonal-vector (vv- second-vector
                                 (sv* projection unit-ref))))
    (if (< (/ (v-length orthogonal-vector)
              (v-length second-vector))
           1e-15)
      (error "input vectors essentially collinear"))
    (unit orthogonal-vector)))
#|
(orthogonal-unit #(1 1 0) #(-2 1 0))
|#

;;; Generate one-axis rotation matrices
;;; Rotate by radians about axis.
(defun rotation-matrix (axis angle)
  "axis angle
Generate a matrix which rotates about the specified axis"
  (let* ((index (cond ((eq axis 'x) 0)
                      ((eq axis 'y) 1)
                      ((eq axis 'z) 2)))
         (sin (sin angle))
         (cos (cos angle))
         (matrix (make-array '(3 4) :element-type 'real
                                    :initial-element 0.0)))
    (setf (aref matrix index index) 1.0)
    (setf (aref matrix
                (mod (+ index 1) 3)
                (mod (+ index 1) 3)) cos)
    (setf (aref matrix
                (mod (+ index 2) 3)
                (mod (+ index 2) 3)) cos)
    (setf (aref matrix
                (mod (+ index 1) 3)
                (mod (+ index 2) 3)) (- sin))
    (setf (aref matrix
                (mod (+ index 2) 3)
                (mod (+ index 1) 3)) sin)
    matrix))

(defun deg->rad (degrees)
  "degrees
conversion from degrees to radians"
  (* degrees 0.0174532925199433))

#|
(rotation-matrix 'x (deg->rad 1.0))
|#






#|

;;; Unused code for general matrix multiplication:
;;; Multiply Mv, where M is a 4 by 4 matrix and v is a vector.
(defun mv* (matrix vector result-vector)
  (dotimes (j 4)
    (let ((rowsum 0.0))
      (dotimes (i 4)
        (incf rowsum (* (aref vector i)
                        (aref matrix j i))))
      (setf (aref result-vector j) rowsum)))
  result-vector)

;;; Multiply MM, where each M is a 4 by 4 matrix.
(defun mm* (left-matrix right-matrix result-matrix)
  (dotimes (k 4)
    (dotimes (j 4)
      (let ((rowsum 0.0))
        (dotimes (i 4)
          (incf rowsum (* (aref right-matrix i k)
                          (aref left-matrix j i))))
        (setf (aref result-matrix k j) rowsum))))
  result-matrix)


(time
 (dotimes (i 1000)
   (mv* #2A((1.0  0.0  0.0  0.0)
            (0.0  1.0  0.0  0.0)
            (0.0  0.0  1.0  0.0)
            (0.0  0.0  0.0  0.0))
        #(1.0  2.0  3.0  4.0)
        #(1.0  2.0  3.0  4.0))
   ))


(time
 (dotimes (i 1000)
   (mm* #2A((2.0  0.0  0.0  10.0)
            (0.0  2.0  0.0  0.0)
            (0.0  0.0  2.0  0.0)
            (0.0  0.0  0.0  2.0))
        #2A((1.0  0.0  0.0  4.0)
            (0.0  1.0  0.0  0.0)
            (0.0  0.0  1.0  0.0)
            (2.0  0.0  0.0  1.0))
        #2A((0.0  0.0  0.0  0.0)
            (0.0  0.0  0.0  0.0)
            (0.0  0.0  0.0  0.0)
            (0.0  0.0  0.0  0.0)))
   ))


|#