;;; A description of how the data structures defined here are used, can
;;; be found in the file Architecture.txt .
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Historical Note: a rough first cut of this file was made by K. E. Drexler.
;;;                  subsequent modifications by KR.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; Atoms are represented by atm structures. The type of the atom is
;;; represented as an integer, used to index into a 2D array of
;;; precomputed values of the squares of the minimum (non-colliding)
;;; distances between atoms of the corresponding types.
;;; Images can be used to represent periodic boundary conditions,
;;; placing virtual atoms wherever a periodic image of the atom
;;; can interact with atoms inside the primary region. Only
;;; an image within the primary region need be tested for clashes.
;;; The clash array takes care of installing and removing images.
;;; (The name "atm" is used because "atom" is a CommonLISP type.)

(defstruct atm
  (xyz #(0.0 0.0 0.0) :type (vector 'real)) ; vector of coordinates,
                                            ;   #(x y z), in pm
  type          ; type (an integer)
  (image nil)   ; pointer to next in a possible chain of images;
                ;    can store a global atom number when walking
                ;    the tree to construct a PDB file
  affiliation)  ; pointer to the moiety of which the atm is part of

;;; Atom type data can be stored in arrays such that
;;; atm-type points to the data associated with an
;;; atom of that type.

;;; Moieties in the moiety-library are represented by
;;; moiety structures. A moiety is linked to its parent
;;; at a boundary defined by the plane of the "crease"
;;; associated with the linking bond(s). Here, atom
;;; coordinates are defined relative to the first three
;;; atoms in the moiety-atm-array, taking the first as
;;; the origin, the second as defining the x axis, and
;;; the third as defining the x-y plane.

(defstruct moiety-type
  number        ; this is the index into an array of moiety-types
  name          ; a unique string naming the moiety
  atm-array     ; an array containing the atoms of the unmodified
                ;   moiety (the first 3 define coordinates)
  bond-array    ; an array with the same ordering as the atm-array,
                ;   each entry containing a list of atm-array indices
                ;   which indicate the other end of the bonds
  sites)        ; array of extension site structures

;;; Most moieties have extension sites, described by site structures.

(defstruct site
  cap-atms      ; list of indices of atms removed when site is used
  root-atms     ; list of indices of atms bonded to when site is used
  link-types)   ; list of possible link types for site,
                ;   ordered according to decreasing volume

;;; A potential relationship between a parent moiety
;;; and a child moiety is represented by a link  structure.

(defstruct link-type
  number        ; this is the index into an array of link-types
  parent        ; key to moiety-type that can be linked to
  parent-site   ; index of link site in parent's moiety and moiety-type arrays
  parent-bonds  ; list of bond-starlets, center-atm as a parent index,
                ;    end-atms as child indices
  child         ; key to moiety-type that can be added
  child-site    ; index of link site in child's moiety and moiety-type arrays
  child-bonds   ; list of bond-starlets, center-atm as a child index,
                ;    end-atms as parent indices
  name          ; string with "parent-name", "child-name", "angle n"
                ;    where "angle n" distinguishes torsional choices
  matrix        ; matrix mapping child atoms into parent coordinates
  volume        ; net added volume: child-volume minus cap-volume
  clash-atms)   ; list of indices of parent atoms that
                ;    clash with child atoms which, as a bonus,
                ;    *includes images explicitly*.

;;; Moieties in the stuffing-tree point to moieties in the
;;; moiety-library to incorporate their generic information, and
;;; contain information specific to their individual states.

(defstruct moiety
  moiety-type   ; pointer to own moiety-type in the library
  parent        ; pointer to parent moiety
  link-type     ; pointer to link-type from parent to this moiety
  (children nil); list of pointers to child moieties
  matrix        ; matrix that maps from moiety to global coordinates
  atm-array     ; copy of moiety-type-atom-array mapped to global
                ;    coordinates
  site-status)  ; array corresponding to site-links in moiety-type,
                ;    elements marked as sites are used or discarded
