;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

(defun attach-if-no-clash-p (parent parent-site-index link-type)
  "parent             ;points to the parent-moiety
   parent-site-index  ;number of the site on parent-moiety
   link-type          ;points to the link-type
Tests if a child can be attached to the parent-moiety at the selected site by
calculating the global coord's of the child and performing clash-detection.
If the child does not clash, the attachment is performed.
The type of the child-moiety plus the attachment orientation are specified by
the chosen link-type.
Returns t if clashed, nil otherwise (including the side-effect of actually
attaching the new moiety, of course)."
  (let* ((child-type (retrieve-from-mlib (link-type-child link-type)))
         (matrix (mm* (moiety-matrix parent) (link-type-matrix link-type)))
         (atm-array (map 'vector
                         #'(lambda (atm)
                             (let ((real-atm (copy-atm atm)))
                               (setf (atm-xyz real-atm)
                                     (mv* matrix (atm-xyz atm)))
                               real-atm))
                         (moiety-type-atm-array child-type)))
         (ignore-list (map 'list #'(lambda (ignore-index)
                                     (elt (moiety-atm-array parent)
                                          ignore-index))
                           (link-type-clash-atms link-type)))
         (child-site-index (link-type-child-site link-type))
         (child-site-cap-atms (map 'list #'(lambda (site-index)
                                             (elt atm-array site-index))
                                   (site-cap-atms
                                    (elt (moiety-type-sites child-type)
                                         child-site-index))))
         (filtered-atm-array (remove-if #'(lambda (atm)
                                           ; (or (eql (atm-type atm) 3)  ;*
                                            (member atm child-site-cap-atms))
                                           ; )                           ;*
                                        atm-array)))
    ;; reverse the filtered-atm-array to let H's clash first
    (if (clash-multiple-p (reverse filtered-atm-array) ignore-list)
;; new moiety has clashed
(progn
;(print "diagnostics-msg from attach-if-no-clash:");*
;(print "  child-to-be-added is:")                 ;*
;(print (moiety-type-name child-type))             ;*
;(print "  ignore-list of parent-moiety is:")      ;*
;(print ignore-list)                               ;*
;(print "  clash-atms-list of child-moiety is:")   ;*
;(print (link-type-clash-atms link-type))
;(print filtered-atm-array)
(print "clashed!")
t)
      ;; new moiety did not clash and so is added now:
      (let* ((number-of-child-sites (length (moiety-type-sites child-type)))
             ;; Construct the newly attached child
             (child (make-moiety
                     :moiety-type     child-type
                     :parent          parent
                     :link-type       link-type
                     :children        (make-array number-of-child-sites
                                                  :initial-element nil)
                     :matrix          matrix
                     :atm-array       atm-array
                     :site-status     (make-array number-of-child-sites
                                                  :initial-element 'open)
                     )))
        (enlist-open-sites child)
        ;; Now modifiy a few things in the parent-moiety to hook it up properly
;(print (moiety-type-name (moiety-moiety-type parent)))
(print (link-type-name link-type))
(print parent-site-index)
        (setf (elt (moiety-children parent) parent-site-index) child)
        (manage-site-status parent parent-site-index 'used)
        (manage-site-status child child-site-index 'used)
        ;; Remove parent's cap-atms from the clash-array
        (delete-multiple-from-clash-array
         (map 'vector
              #'(lambda (site-cap-index)
                  (elt (moiety-atm-array parent)
                       site-cap-index))
              (site-cap-atms (elt
                              (moiety-type-sites (moiety-moiety-type parent))
                              parent-site-index))))
        ;; Enter child-atms into the clash-array
        (insert-multiple-into-clash-array
         (affiliate-multiple-atms-with-moiety filtered-atm-array child))
        
        (incr-tree-volume-by (link-type-volume link-type))
        (values nil child)) ;; the return-value for not having clashed
       );;if clash
    ))

;;; following routine not yet tested:
(defun detach-moiety (parent site)
  "parent       ;points to the parent-moiety
   site         ;number of the site on parent-moiety
Detaches a child-moiety (including that whole subtree) from a selected
parent-moiety at the selected site."
  (let* ((subtree (elt (moiety-children parent) site)))
    ;; Now modifiy a few things in the parent-moiety to delete references to
    ;;   the subtree
    (setf (elt (moiety-children parent) site) nil)
    (setf (elt (moiety-site-status parent) site) 'open)

    ;; insert here the setting of all 'blocked sites to 'open

    (incr-tree-volume-by (- (erase-subtree subtree)))
    ))

(defun incr-tree-volume-by (volume-increment)
  "volume-increment
Increases the volume of the currently active stuffing-tree
by the supplied value"
  (set-current-tree-volume (+ volume-increment (get-current-tree-volume))))

(defun init-tree (frame-moiety)
  "frame-moiety   ;string which is name of a moiety-type
Constructs the root of a tree and sets the tree-volume to zero.
Any kind of moiety-type could serve as the root, but often it will be a
predefined frame-work which constitutes a cell cut out of a lattice which
surrounds a cavity to be stuffed."
  (let* ((frame-moiety-type (retrieve-from-mlib frame-moiety))
         (number-of-frame-sites (length (moiety-type-sites frame-moiety-type)))
         ;; Construct the root-moiety
         (tree-root (make-moiety
                     :moiety-type     frame-moiety-type
                     :parent          nil
                     :link-type       nil
                     :children        (make-array number-of-frame-sites
                                                  :initial-element nil)
                     :matrix          (new-unit-matrix)
                     :atm-array       (moiety-type-atm-array frame-moiety-type)
                     :site-status     (make-array number-of-frame-sites
                                                  :initial-element 'open)
                     )))
    (set-clash-environment (list (list -1500 -1500 -1500) (list 1500 1500 1500))
                           (atm-MM2-0.1nN-radii))

    (setf Current-tree (gensym))
    (set-current-tree tree-root)
    (init-current-tree-volume)
    (init-open-sites)
    (enlist-open-sites tree-root)

    (insert-multiple-into-clash-array
     (affiliate-multiple-atms-with-moiety (moiety-atm-array tree-root)
                                          tree-root))
    (get-current-tree)
    ))

(defun ignore-with-images (primary-atm-ignore-list)
  "primary-atm-ignore-list
Appends to the primary-atm-ignore-list all the images of these atms,
so they too will be ignored in the clash-array."
  (labels ((list-images
            (atm)
            (if (null (atm-image atm))
              (list atm)
              (cons atm (list-images (atm-image atm))))))
    (apply #'append (mapcar #'(lambda (atm) (list-images atm))
                            primary-atm-ignore-list))))

(defun affiliate-atm-with-moiety (atm moiety)
  "atm
   moiety
Affiliates the atm with the moiety of which it should be part of, so that this
moiety can be identified again starting from a known atm in the clash-array.
This is a forcible & destructive operation."
  (setf (atm-affiliation atm) moiety))

(defun affiliate-multiple-atms-with-moiety (atm-array moiety)
  "atm-array
   moiety
Affiliates the atms with the moiety of which they should be part of, so that
this moiety can be identified again starting from a known atm in the
clash-array. This is a forcible & destructive operation and returns the
modified atm-array."
(dosequence (atm atm-array atm-array)
              (affiliate-atm-with-moiety atm moiety)))


