;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@shell.portal.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; General commentary regarding mechanism: see at the bottom of this file.

(defun dendro-stenogen (input-tree)
  "input-tree
Distills a tree structure into its slim shorthand representation, based on
the unique identifiers that the link-types have in the moiety-library."
  (let ((node-mapper
         (make-node-mapping-info
          :input-node-topology-list
              #'(lambda (input-tree entry-index)
                  (let ((result-tree-node (make-array 0 :adjustable t
                                                      :fill-pointer 0)))
                    (coerce       ;; convert the vector to a list
                     (doseq-with-index (branch-status index
                                        (moiety-site-status input-tree)
                                        result-tree-node) ; return the vector
                       (vector-push-extend
                        (if (equal index entry-index)
                          ;; entry-site, mark specially
                          (list 'entry nil nil nil)
                          ;; regular branch-site
                          (if (equal branch-status 'used)
                            ;; branch-status is 'used
                            (let* ((actual-branch (elt
                                                   (moiety-children input-tree)
                                                   index))
                                   (branch-link-type (moiety-link-type
                                                      actual-branch)))
                              (list
                               'branch
                               actual-branch         ; points to the subtree
                               (link-type-name       ; link-id for subtree
                                branch-link-type)
                               (link-type-child-site ; entry-index into
                                branch-link-type)))  ;   the lower subtree
                            ;; branch-status is 'open
                            (list 'leaf nil
                                  nil              ;no supplementary info
                                  nil))            ;
                          )
                        result-tree-node))
                     'list)))
          :map-input-branch     #'(lambda (branch) branch)
          :map-input-leaf       #'(lambda (leaf) (declare (ignore leaf)) 'cap)
          :empty-result-node    #'(lambda (input-node-topology entry-index-dummy)
                                    (declare (ignore entry-index-dummy))
                                    (let ((result
                                           (make-list
                                            (length input-node-topology)))
                                          (navel-index
                                           (position-if
                                            #'(lambda (site-info)
                                                (eql (first site-info)
                                                     'entry))
                                            input-node-topology)))
                                      ;; the implicit assumption here is that
                                      ;;   there is no more than one 'navel-site
                                      ;;   (no error-checks are performed)
                                      (unless (equal navel-index nil)
                                        (setf (elt result
                                                   navel-index)
                                              'navel))
                                      result))
          :enter-output-branch  #'(lambda (branch-result result-tree-node 
                                           index link-id)
                                    (setf (elt result-tree-node index)
                                          (list link-id
                                                branch-result)))
          :enter-output-leaf    #'(lambda (leaf-result result-tree-node
                                           index dummy)
                                    (declare (ignore dummy))
                                    (setf (elt result-tree-node index)
                                          leaf-result))))
        )
    ;; Is this tree rooted in a framework ? Print it if yes.
    (if (eql (moiety-parent input-tree) nil)
      (list
       (list 'frame (moiety-type-name (moiety-moiety-type input-tree)))
       ;; invoke walking of bi-directional tree by supplying the third argument
       (tree-traverse input-tree node-mapper nil))

      ;; invoke walking of bi-directional tree by supplying the third argument
      (tree-traverse input-tree node-mapper nil))
    ))

(defun dendro-plerogen (input-tree)
  "input-tree
Expands the shorthand representation of a tree back to its full beauty and
bountyful structural complexity (:-)"
  (let ((node-mapper
         (make-node-mapping-info
          :input-node-topology-list
              #'(lambda (input-tree dummy)
                  (declare (ignore dummy))
                  (map 'list #'(lambda (site)
                                 (if (listp site)
                                   (list
                                    'branch
                                    (second site)  ; the subtree
                                    (first site)   ; link-id to subtree
                                    nil)           ; to be filled in later!
                                   (ecase site
                                     (cap   (list 'leaf nil nil nil))
                                     (navel (list 'entry nil nil nil))
                                     )))
                       input-tree))
          :map-input-branch     #'(lambda (branch) branch)
          :map-input-leaf       #'(lambda (leaf) leaf)
          :empty-result-node    ;; executed only for the side-effects
              #'(lambda (input-node-topology parent)
                  ;; also fills-in the parent pointers for the sub-branches
                  (doseq-with-index (branch-site index input-node-topology)
                    (when (eql (first branch-site) 'branch)
                      (multiple-value-bind
                        (clash? node)
                        (attach-if-no-clash-p
                         parent
                         index
                         (retrieve-from-mlib (third branch-site)))
                        (if clash?
                          (error "should NOT clash !!")
                          (setf (fourth branch-site) node))
                        ))))
          :enter-output-branch  #'(lambda (branch-result result-tree-node 
                                           index link-id)
                                    (declare (ignore branch-result 
                                                     result-tree-node
                                                     index link-id)))
          :enter-output-leaf    #'(lambda (leaf-result result-tree-node
                                           index dummy)
                                    (declare (ignore leaf-result 
                                                     result-tree-node
                                                     index dummy)))
          )))
    ;; Is this tree rooted in a framework ?
    (if (eql (first (first input-tree)) 'frame)
      (let
        ;; initializes the clash-array and other stuff
        ((init-node (init-tree (second (first input-tree)))))
        ;; invoke walking of bi-directional tree by supplying the third argument
        (tree-traverse (second input-tree) node-mapper init-node))
      (error "no framework specified."))
    ))

;;; The structure of a node in the compressed tree is a list which contains
;;; all the sites of the nodes. These sites can either have the value 'cap
;;; to denote a terminated leaf, or they can be a list, the first element
;;; of which is the link-type leading to the subtree, and the second of which
;;; is again a list, which is the first node of that subtree.

(defun tree-traverse (input-tree
                        node-map-info
                        &optional (entry-index nil entry-supplied-p))
  "input-tree         ;points to the root of the tree (or subtree)
   node-map-info      ;structure that contains info on how to access the 
                      ; branches leading away from a node in the input-tree, etc.
   &optional entry-index ; needed for bi-directional trees
Recursively walks through the given tree structure in depth-first fashion while
simultaneously constructing the tree structure that is returned as a result."
  (if entry-supplied-p
    ;;; go down on a bi-directional tree
    (let* ((topology-list (funcall (node-mapping-info-input-node-topology-list 
                                    node-map-info)
                                   input-tree
                                   entry-index))
           (result-tree-node (funcall (node-mapping-info-empty-result-node 
                                       node-map-info)
                                      topology-list
                                      entry-index)))
      (doseq-with-index (branch-site index topology-list result-tree-node)
        (ecase (first branch-site)
          (branch (funcall (node-mapping-info-enter-output-branch 
                            node-map-info)
  ;; recursive descent occurs here:
                           (tree-traverse
                            (funcall (node-mapping-info-map-input-branch 
                                      node-map-info)
                                     (second branch-site))
                            node-map-info
                            (fourth branch-site)) ;; entry-index for subtree
                           result-tree-node
                           index
                           (third branch-site)))
          (leaf (funcall (node-mapping-info-enter-output-leaf 
                          node-map-info)
                         (funcall (node-mapping-info-map-input-leaf 
                                   node-map-info)
                                  (second branch-site))
                         result-tree-node
                         index
                         (third branch-site)))
          ;; don't follow up on the entry-site: that's where you came from!
          (entry nil)    ;; do nothing: that's where you came from!
          ))
      )
    ;;; go down on a regular, strictly hierarchical tree
    (let* ((topology-list (funcall (node-mapping-info-input-node-topology-list 
                                    node-map-info)
                                   input-tree))
           (result-tree-node (funcall (node-mapping-info-empty-result-node 
                                       node-map-info)
                                      topology-list)))
      (doseq-with-index (branch-site index topology-list result-tree-node)
        (ecase (first branch-site)
          (branch (funcall (node-mapping-info-enter-output-branch 
                             node-map-info)
  ;; recursive descent occurs here:
                            (tree-traverse
                             (funcall (node-mapping-info-map-input-branch 
                                       node-map-info)
                                      (second branch-site))
                             node-map-info)
                            result-tree-node
                            index
                            (third branch-site)))
          (leaf (funcall (node-mapping-info-enter-output-leaf 
                           node-map-info)
                          (funcall (node-mapping-info-map-input-leaf 
                                    node-map-info)
                                   (second branch-site))
                          result-tree-node
                          index
                          (third branch-site)))
          (entry nil)    ;; do nothing: this is only a placeholder in strictly
          ))             ;;   hierarchical trees, to keep the indices straight
      )
    ))

(defstruct node-mapping-info
  input-node-topology-list  ; function that returns the list of branches 
                            ;   and leaves.
                            ; Usually, has one parameter: (input-node)
                            ; With bi-directional trees, has two parameters:
                            ;   (input-node entry-index)
  map-input-branch          ; function for branch mapping. Has one parameter
  map-input-leaf            ; function for leaf mapping. Has one parameter
  empty-result-node         ; function which produces the empty output node
                            ;   to be returned after filling in stuff
                            ; Usually, has one parameter: (input-node-topology)
                            ; With bi-directional trees, has two parameters:
                            ;   (input-node-topology entry-index)
  enter-output-branch       ; function for branch mapping. Has 4 parameters:
                            ;   (branch-result result-tree-node index supplement)
  enter-output-leaf         ; function for leaf mapping. Has 4 parameters:
                            ;   (leaf-result result-tree-node index supplement)
)

;;; The topology-list is a list of lists with three items, first a flag 
;;; indicating 'branch or 'leaf, and second has to come the pointer to the 
;;; branch or leaf. In the case of a branch, this will point to the appropriate 
;;; subtree. The third item is supplementary information, which is specific
;;; to the implementation of various tree-types.
;;; For "bi-directional" trees, there is an additional kind of flag, 'entry,
;;; which indicates that the so marked-up branch-site should not be followed,
;;; because it will lead back precisely to the place from which one entered into
;;; the current tree-node. In addition, each site-describing list in the
;;; topology-list has a fourth item, which is currently only used concurrently
;;; with the flag 'branch, and it denotes the entry-index, the number of the
;;; entry-site for the tree-node in the respective subtree. This information is
;;; needed in the next lower level of node-processing, to avoid the problem just
;;; described.
